<?php

namespace Drupal\sqrl\Form;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Entity\EntityStorageException;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Mail\MailManagerInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\sqrl\Entity\Identity;
use Drupal\sqrl\Sqrl;
use Drupal\sqrl\StringManipulation;
use Drupal\user\Entity\User;
use Drupal\user\UserDataInterface;
use Drupal\user\UserInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a SQRL form.
 */
class ProfileEdit extends FormBase {

  use StringManipulation;

  /**
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  protected $currentUser;

  /**
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * @var \Drupal\sqrl\Nut
   */
  protected $nut;

  /**
   * @var \Drupal\Core\Mail\MailManagerInterface
   */
  protected $mailManager;

  /**
   * @var \Drupal\user\UserDataInterface
   */
  protected $userData;

  /**
   * Link constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   * @param \Drupal\sqrl\Sqrl $sqrl
   * @param \Drupal\Core\Session\AccountProxyInterface $current_user
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   * @param \Drupal\Core\Mail\MailManagerInterface $mail_manager
   * @param \Drupal\user\UserDataInterface $user_data
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, Sqrl $sqrl, AccountProxyInterface $current_user, MessengerInterface $messenger, MailManagerInterface $mail_manager, UserDataInterface $user_data) {
    $this->entityTypeManager = $entity_type_manager;
    $this->currentUser = $current_user;
    $this->messenger = $messenger;
    $this->mailManager = $mail_manager;
    $this->userData = $user_data;

    $this->nut = $sqrl->getNewNut();
    $this->nut->fetch();
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): ProfileEdit {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('sqrl.handler'),
      $container->get('current_user'),
      $container->get('messenger'),
      $container->get('plugin.manager.mail'),
      $container->get('user.data')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId(): string {
    return 'sqrl_profile_edit';
  }

  /**
   * @param \Drupal\user\UserInterface $user
   * @param string $token
   *
   * @return \Drupal\Core\Access\AccessResult
   */
  public function access(UserInterface $user, $token): AccessResult {
    if ($this->nut->isValid() && $this->nut->getLoginToken() === $token && $this->currentUser->id() === $user->id()) {
      return AccessResult::allowed();
    }
    return AccessResult::forbidden();
  }

  /**
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   * @param \Drupal\user\UserInterface $user
   * @param string $token
   *
   * @return array
   */
  public function buildForm(array $form, FormStateInterface $form_state, $user = NULL, $token = NULL): array {
    $form['pass'] = [
      '#type' => 'password_confirm',
      '#size' => 25,
      '#description' => $this->t('To set your user password, enter the password in both fields.'),
    ];
    $form['mail'] = [
      '#type' => 'email',
      '#title' => $this->t('Email address'),
      '#description' => $this->t('A valid email address. All emails from the system will be sent to this address. The email address is not made public and will only be used if you wish to receive a new password or wish to receive certain news or notifications by email.'),
      '#default_value' => (strpos($user->getEmail(), Identity::DEFAULT_EMAIL_DOMAIN) === FALSE) ? $user->getEmail() : '',
    ];

    $form['#validate'][] = [$this, 'validateMail'];

    $form['actions'] = [
      '#type' => 'actions',
    ];
    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Save'),
    ];

    return $form;
  }

  /**
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @noinspection PhpUnusedParameterInspection
   */
  public function validateMail($form, FormStateInterface $form_state): void {
    $mail = trim($form_state->getValue('mail'));
    $users = $this->entityTypeManager->getStorage('user')->loadByProperties(['mail' => $mail]);
    foreach ($users as $user) {
      if ($user->id() !== $this->currentUser->id()) {
        $form_state->setErrorByName('mail', $this->t('Mail address already taken.'));
        return;
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {
    $pass = trim($form_state->getValue('pass'));
    $mail = trim($form_state->getValue('mail'));

    /** @var \Drupal\user\UserInterface $account */
    $account = User::load($this->currentUser->id());
    $changed = FALSE;
    if (!empty($mail) && $mail !== $account->getEmail()) {
      if ($this->configFactory()->get('sqrl.settings')->get('mail_verification')) {
        $token = $this->base64_encode($this->randomBytes(8));
        // Store mail address.
        $this->userData->set('sqrl', $this->currentUser->id(), $token, $mail);
        // Send out email.
        $this->mailManager->mail('sqrl', 'confirm_email', $mail, $this->currentUser->getPreferredLangcode(), [
          'account' => $this->currentUser,
          'token' => $token,
        ]);
        $this->messenger->addStatus($this->t('We have sent you an email with a link to confirm your new email address.'));
      }
      else {
        $account->setEmail($mail);
        $changed = TRUE;
      }
    }
    if (!empty($pass)) {
      $account->setPassword($pass);
      $changed = TRUE;
    }
    if ($changed) {
      try {
        $account->save();
      }
      catch (EntityStorageException $e) {
        // TODO: Log this exception.
      }
    }

    // Redirect back to the user profile.
    $form_state->setRedirect('user.page');
  }

}
